%{

#include <stdio.h>
#define KRB5_CONFIG "/etc/krb5.conf"
#ifdef KRB5CONF_APP
#define YYDEBUG 1
#endif

int xkrb5_conf_lineno = 1;
char *xkrb5_conf_section = NULL;
struct xkrb5_conf_entry {
	char *key;
	char *value;
	struct xkrb5_conf_entry *next;
} *xkrb5_conf_entries = NULL;
int xkrb5_conf_parse(void);
int xkrb5_conf_lex(void);
extern char *xkrb5_conf_text;
extern FILE *xkrb5_conf_in;

%}

%union {
	char *sval;
	int ival;
}

%token <sval> STRING
%token <sval> KEY
%type  <sval> strings
%token <sval> EQUALS
%token NEWLINE
%token <sval> SECTION
%token CURLYLEFT
%token CURLYRIGHT

%%

lines:
	line |
	lines line;
line:
	NEWLINE |
	sectionstart |
	assignment |
	subsection;

sectionstart:
	SECTION NEWLINE
	{
		if(xkrb5_conf_section) {
			free(xkrb5_conf_section);
		}
		xkrb5_conf_section = $1;
	};

subsection:
	subsectionstart subsectionstop |
	subsectionstart lines subsectionstop;

subsectionstart:
	KEY EQUALS CURLYLEFT NEWLINE
	{
		char *tmp;
		if(xkrb5_conf_section) {
			tmp = malloc(strlen(xkrb5_conf_section) + 1 + strlen($1) + 1);
			strcpy(tmp, xkrb5_conf_section);
			strcat(tmp, "\177");
			strcat(tmp, $1);
			free(xkrb5_conf_section);
			xkrb5_conf_section = tmp;
		}
	};

subsectionstop:
	CURLYRIGHT NEWLINE
	{
		if(xkrb5_conf_section) {
			char *p;
			p = strrchr(xkrb5_conf_section, '\177');
			if(p) {
				*p = '\0';
			}
		}
	};

assignments:
	assignment |
	assignments assignment;

assignment:
	NEWLINE |
	KEY EQUALS strings NEWLINE
	{
		struct xkrb5_conf_entry *entry = NULL;
		size_t l;
		entry = malloc(sizeof(struct xkrb5_conf_entry));
		memset(entry, 0, sizeof(struct xkrb5_conf_entry));
		l = strlen(xkrb5_conf_section) + 1 + strlen($1) + 1;
		entry->key = malloc(l);
		strcpy(entry->key, xkrb5_conf_section);
		strcat(entry->key, "\177");
		strcat(entry->key, $1);
		entry->value = strdup($3);
		entry->next = xkrb5_conf_entries;
		xkrb5_conf_entries = entry;
	};

strings:
	KEY
	{
		$$ = strdup($1);
	}
	| strings KEY
	{
		char *tmp;
		tmp = malloc(strlen($$) + strlen($2 ?: "") + 2);
		strcpy(tmp, $$);
		strcat(tmp, " ");
		strcat(tmp, $2 ?: "");
		$$ = tmp;
	}
	| strings EQUALS KEY
	{
		char *tmp;
		tmp = malloc(strlen($$) + strlen($3 ?: "") + 2);
		strcpy(tmp, $$);
		strcat(tmp, "=");
		strcat(tmp, $3 ?: "");
		$$ = tmp;
	};
%%

#ifdef KRB5CONF_APP
#ifndef APPDEFAULT_APP
#define APPDEFAULT_APP "pam"
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

#ifndef CRIT
#define CRIT printf
#endif
#endif

int
yyerror(const char *error)
{
	CRIT("error parsing " KRB5_CONFIG " at line %d at `%s': %s\n",
	     xkrb5_conf_lineno, xkrb5_conf_text, error);
	return 0;
}

static int
yywrap(void)
{
	return 1;
}

const char *
xkrb5_conf_read(const char *key)
{
	struct xkrb5_conf_entry *entry;
	char buf[8192];

	snprintf(buf, sizeof(buf), "%s\177%s",
		 APPDEFAULT_APP, key);
	for(entry = xkrb5_conf_entries;
	    entry != NULL;
	    entry = entry->next) {
		if(strcmp(entry->key, buf) == 0) {
			return entry->value;
		}
	}
#ifndef HAVE_KRB5_APPDEFAULT_STRING
	snprintf(buf, sizeof(buf), "%s\177%s\177%s",
		 "appdefaults", APPDEFAULT_APP, key);
	for(entry = xkrb5_conf_entries;
	    entry != NULL;
	    entry = entry->next) {
		if(strcmp(entry->key, buf) == 0) {
			return entry->value;
		}
	}
#endif
	return NULL;
}

void
xkrb5_conf_parse_file(void)
{
	static int parsed = FALSE;
#ifdef KRB5CONF_APP
	xkrb5_conf_debug = TRUE;
#endif
	if(!parsed) {
		xkrb5_conf_in = fopen(KRB5_CONFIG, "r");
		if(xkrb5_conf_in) {
			do {
				xkrb5_conf_parse();
			} while(!feof(xkrb5_conf_in));
			fclose(xkrb5_conf_in);
		}
		parsed = TRUE;
	}
}

#ifdef KRB5CONF_APP
int
main(int argc, char **argv)
{
	struct xkrb5_conf_entry *entry;
	xkrb5_conf_parse_file();
	for (entry = xkrb5_conf_entries;
	     entry != NULL;
	     entry = entry->next) {
		printf("`%s' = `%s'\n", entry->key, entry->value);
	}
	return 0;
}
#endif
